import numpy as np
import sasktran as sk
from tqdm import tqdm
import skdoas.amf.amf
import skdoas.wf.do_wf

from typing import Union, List


class AirMassFactorBaseDO(skdoas.amf.amf.AirMassFactorBase):
    """ 
    Base class for objects that calculate air mass factors using SASKTRAN's discrete ordinates engine.
    
    See the base class for more documentation.
    """
    def __init__(self, geometry: sk.Geometry, atmosphere: sk.Atmosphere, amf_species: str,
                 box_boundaries: np.array, box_midpoints: Union[None, np.array]=None,
                 wf_widths: Union[None, np.array]=None, total_sum: bool=False, total_remove: bool=False, **options):
        super().__init__(geometry, atmosphere, amf_species, box_boundaries, box_midpoints, wf_widths, total_sum,
                         total_remove, **options)

    def _set_boxes(self):
        super()._set_boxes()
        self.options['wfaltitudes'] = self.box_midpoints

    @property
    def reference_points(self):
        """
        numpy.array : Returns the reference points corresponding to each line of sight in self.geometry. The array has
        shape (len(self.geometry.lines_of_sight), 4), and each entry contains latitude, longitude, altitude and mjd 
        (in that order).
        """
        # calculate surface_height intersections TODO add ref point customization and/or altitude control
        los = self.lines_of_sight

        ref_pts = np.zeros((len(los), 4))
        for i, s in enumerate(los):
            ref = s.ground_intersection()
            if ref is None:
                ref = s.tangent_location()
            ref_pts[i, :] = [ref.latitude, ref.longitude, ref.altitude, s.mjd]
        return ref_pts


class AirMassFactorDO(AirMassFactorBaseDO):
    """ 
    Calculate AMFs using SASKTRAN's discrete ordinates engine's built-in weighting function algorithm. 
    
    See base classes for more documentation.
    """
    
    def __init__(self, geometry: sk.Geometry, atmosphere: sk.Atmosphere, amf_species: str,
                 box_boundaries: np.array, box_midpoints: Union[None, np.array]=None,
                 wf_widths: Union[None, np.array]=None, total_sum: bool=False, total_remove: bool=False, **options):
        """ For parameter definitions see Ai rMassFactorWFBase.__init__. """
        self._wf_obj = skdoas.wf.do_wf.WeightingFunctionDO(geometry, atmosphere, amf_species, **options)
        super().__init__(geometry, atmosphere, amf_species, box_boundaries, box_midpoints, wf_widths, total_sum,
                         total_remove, **options)


class AirMassFactorFiniteDifferenceDO(AirMassFactorBaseDO, skdoas.amf.amf.AirMassFactorFiniteDifferenceBase):
    """ 
    Calculate AMFs using SASKTRAN's discrete ordinates engine to evaluate finite-difference weighting functions. 
    
    Use of this class is not recommended as AirMassFactorDO will produce equivalent results in a fraction of the time.
    
    See base classes for more documentation.
    """
    def __init__(self, geometry: sk.Geometry, atmosphere: sk.Atmosphere, amf_species: str,
                 box_boundaries: np.array, box_midpoints: Union[None, np.array]=None,
                 wf_widths: Union[None, np.array]=None, total_sum: bool=False, total_remove: bool=False,
                 method: str='forward', step: Union[float, List[float]]=0.1, relative_step: bool=True,
                 minimum_step: float=1e8, progress_bar: Union[bool, tqdm]=False, **options):
        """ For parameter definitions see AirMassFactorFiniteDifferenceBase.__init__. """
        self._wf_obj = skdoas.wf.do_wf.WeightingFunctionFiniteDifferenceDO(
            geometry, atmosphere, amf_species, method, step, relative_step, minimum_step, progress_bar, **options
        )
        AirMassFactorBaseDO.__init__(
            self, geometry, atmosphere, amf_species, box_boundaries, box_midpoints, wf_widths, total_sum, total_remove,
            **options
        )
        skdoas.amf.amf.AirMassFactorFiniteDifferenceBase.__init__(
            self, geometry, atmosphere, amf_species, box_boundaries, box_midpoints, wf_widths, total_sum, total_remove,
            method, step, relative_step, minimum_step, progress_bar, **options
        )
