import numpy as np
import sasktran as sk
import skdoas.amf.amf
import skdoas.wf.mc_wf


class AirMassFactorMC(skdoas.amf.amf.AirMassFactorBase):
    """
    Calculate AMFs using SASKTRAN's Monte Carlo engine. AMFs are calculated directly by tracing individual
    photon paths within each AMF layer, as opposed to the other engines which use weighting functions.

    This class currently requires a non-standard branch of the underlying SASKTRAN framework to be installed.

    See the base class for more documentation.
    """
    def __init__(self, geometry: sk.Geometry, atmosphere: sk.Atmosphere, amf_species: str,
                 box_boundaries: np.array, total_sum: bool=False, total_remove: bool=False,
                 **options):
        """ For parameter definitions see AirMassFactorBase.__init__. """
        self._wf_obj = skdoas.wf.mc_wf.WeightingFunctionNullMC(geometry, atmosphere, **options)
        super().__init__(geometry, atmosphere, amf_species, box_boundaries, total_sum, total_remove, **options)
        self.options['setminfractionhigherorder'] = 1.0  # turns off high order truncation optimization, necessary for AMF calculations for now

    @property
    def box_boundaries(self):
        return self._box_boundaries

    @box_boundaries.setter
    def box_boundaries(self, value: np.array):
        self._box_boundaries = value
        self.options['manualamfshells'] = value

    @property
    def reference_points(self):
        # fetch reference point
        ref_pts = np.zeros((1, 4))
        engine = sk.EngineHR(self.geometry, self.atmosphere, [500.0], self.options)
        engine._add_lines_of_sight()
        engine._initialize_model()
        ref_pts[0, :] = engine.model_parameters['referencepoint']
        return ref_pts

    def __call__(self, wl):
        # Get dimensions
        nwl = len(wl)
        nlos = len(self.lines_of_sight)
        nbox = len(self.box_boundaries) - 1
        box_midpoints = 0.5 * (self.box_boundaries[:-1] + self.box_boundaries[1:])

        rad = self.radiances(wl)
        amfs = np.swapaxes(self._wf_obj.engine.skif_object().GetProperty('airmassfactor')[1].reshape(nlos, nwl, nbox, order='f'), 0, 1)
        amfvars = np.swapaxes(self._wf_obj.engine.skif_object().GetProperty('airmassfactorvariance')[1].reshape(nlos, nwl, nbox, order='f'), 0, 1)
        rad['amf'] = (['wavelength', 'los', 'altitude'], amfs)
        rad['amfstd'] = (['wavelength', 'los', 'altitude'], np.sqrt(amfvars))

        rad.coords['altitude'] = ('altitude', box_midpoints)
        rad = rad.stack(perturbation=('altitude',))

        # calculate the total AMFs
        if self.total_sum:
            self._calc_total_amf_sum(wl, rad)
        if self._total_remove:
            self._calc_total_amf_remove_absorber(wl, rad)

        return rad