import numpy as np
import matplotlib.pyplot as plt
import skdoas.ring.rrs_cross_sections


class RingSpectrum(object):
    """
    Calculates the ring spectrum from rotational Raman
    scattering of O2 and N2 particles. For use in DOAS retrievals.

    Parameters
    ----------
    solar_spectrum_wavelength : np.ndarray
        Array of wavelengths corresponding to the irradiance data.
    solar_spectrum_irradiance : np.ndarray
        Solar top-of-atmosphere irradiance data.
    """
    def __init__(self, solar_spectrum_wavelength, solar_spectrum_irradiance):
        self._wl = solar_spectrum_wavelength
        self._irr = solar_spectrum_irradiance

        self.n2_levels = 25
        self.o2_levels = 25
        self.temperature = 250.0
        self.full = False
        self.polynomial_order = 3
        self.polynomial_removal = 'subtract'

        self._last_wl = None
        self._last_source = None

    @property
    def n2_levels(self) -> int:
        """
        int : Number of N2 energy levels to consider in the calculation.
        Default: 25.
        """
        return self._n2_levels

    @n2_levels.setter
    def n2_levels(self, value):
        self._n2_levels = value

    @property
    def o2_levels(self):
        """
        int : Number of O2 energy levels to consider in the calculation.
        Default: 25.
        """
        return self._o2_levels

    @o2_levels.setter
    def o2_levels(self, value):
        self._o2_levels = value

    @property
    def temperature(self) -> float:
        """
        float : Effective temperature.
        Default: 250K.
        """
        return self._temp

    @temperature.setter
    def temperature(self, value):
        self._temp = value

    @property
    def full(self) -> bool:
        """
        bool : True to calculate the true RRS cross sections at each individual wavelength for convolution. False to
        use cross sections from a single incident wavelength as suggested by Han et. al. Default: False.
        """
        return self._full

    @full.setter
    def full(self, value):
        self._full = value

    @property
    def polynomial_order(self) -> int:
        """
        int : The order of the polynomial to remove from the final spectrum.
        Default: 3.
        """
        return self._poly_order

    @polynomial_order.setter
    def polynomial_order(self, value):
        self._poly_order = value

    @property
    def polynomial_removal(self) -> str:
        """
        str : The method used to remove the polynomial from the final spectrum. Should be one of 'subtract', 'divide',
        or 'None'. Default: 'subtract'.
        """
        return self._poly_remove

    @polynomial_removal.setter
    def polynomial_removal(self, value):
        if isinstance(value, str):
            if value.lower() == 'subtract':
                self._poly_remove = 'subtract'
            elif value.lower() == 'divide':
                self._poly_remove = 'divide'
            else:
                self._poly_remove = None
        else:
            self._poly_remove = None

    def source_spectrum(self, wavelength, last_source=False):
        """
        Calculates the Ring effect source spectrum from rotational Raman scattering of O2 and N2 particles.

        Parameters
        ----------
        wavelength : np.ndarray
            The wavelengths for the ring spectrum calculation. If these wavelengths do not fall within the wavelength
            range of the irradiance data, results will be inaccurate.
        last source : bool
            True to reuse the source spectrum cached in the last call to either source_spectrum() or
            differential_ring_spectrum(). False to calculate a new source spectrum. If there is no cached spectrum, or
            if the current wavelength does not match the wavelength used for the cached spectrum, a new spectrum is
            calculated whether last_source is True or False.

        Returns
        -------
        np.ndarray
            Array containing the source spectrum.
        """

        if last_source and np.all(wavelength == self._last_wl) and self._last_source is not None:
            return self._last_source

        if self.full:
            self._last_wl = wavelength
            self._last_source = np.zeros_like(wavelength)
            for i, wl in enumerate(wavelength):
                rrso2 = skdoas.ring.rrs_cross_sections.RRSCrossSectionsO2(wl)
                rrsn2 = skdoas.ring.rrs_cross_sections.RRSCrossSectionsN2(wl)
                wlo2, cso2 = rrso2.calc_cross_sections(self._o2_levels, self._temp)
                wln2, csn2 = rrsn2.calc_cross_sections(self._n2_levels, self._temp)
                if len(wlo2) > len(wln2):
                    wlh = wlo2
                    csn2 = np.interp(wlh, wln2, csn2, left=0.0, right=0.0)
                else:
                    wlh = wln2
                    cso2 = np.interp(wlh, wlo2, cso2, left=0.0, right=0.0)
                csatmo = 0.21 * cso2 + 0.78 * csn2
                irrh = np.interp(wlh, self._wl, self._irr)
                self._last_source[i] = np.convolve(irrh, csatmo, 'valid')[0]
        else:
            centre_wavelength = 0.5 * (wavelength[0] + wavelength[-1])
            rrso2 = skdoas.ring.rrs_cross_sections.RRSCrossSectionsO2(centre_wavelength)
            rrsn2 = skdoas.ring.rrs_cross_sections.RRSCrossSectionsN2(centre_wavelength)
            wlo2, cso2 = rrso2.calc_cross_sections(self._o2_levels, self._temp)
            wln2, csn2 = rrsn2.calc_cross_sections(self._n2_levels, self._temp)

            # get the width of the ring effect
            width = max([wlo2[-1] - wlo2[0], wln2[-1] - wln2[0]])
            extend = round(width / 2, 3) + 0.001

            # zero-pad the cross-sections to match the window
            wlh = np.arange(wavelength[0] - extend, wavelength[-1] + extend + 0.001, 0.001)
            csn2 = np.interp(wlh, wln2, csn2, left=0.0, right=0.0)
            cso2 = np.interp(wlh, wlo2, cso2, left=0.0, right=0.0)
            csair = 0.21 * cso2 + 0.78 * csn2

            irrh = np.interp(wlh, self._wl, self._irr)
            self._last_wl = wavelength
            self._last_source = np.convolve(irrh, csair, 'same')
            self._last_source = np.interp(wavelength, wlh, self._last_source)

        return self._last_source

    def differential_ring_spectrum(self, wavelength, last_source=False):
        """
        Calculates the differential ring spectrum from rotational Raman scattering of O2 and N2 particles. For use in
        DOAS retrievals.

        Parameters
        ----------
        wavelength : np.ndarray
            The wavelengths for the ring spectrum calculation. If these wavelengths do not fall within the wavelength
            range of the irradiance data, results will be inaccurate.
        last source : bool
            True to reuse the source spectrum cached in the last call to either source_spectrum() or
            differential_ring_spectrum(). False to calculate a new source spectrum. If there is no cached spectrum, or
            if the current wavelength does not match the wavelength used for the cached spectrum, a new spectrum is
            calculated whether last_source is True or False.

        Returns
        -------
        np.ndarray
            Array containing the differential ring spectrum.
        """

        if not last_source or np.any(wavelength != self._last_wl) or self._last_source is None:
            # get the source spectrum
            self._last_wl = wavelength
            self._last_source = self.source_spectrum(wavelength)

        # divide out the original solar spectrum to get the differential spectrum
        diff = self._last_source / np.interp(wavelength, self._wl, self._irr)

        # fit a polynomial to remove
        if self.polynomial_removal is 'subtract':
            p = np.polyfit(wavelength, diff, self.polynomial_order)
            diff_no_poly = diff - np.polyval(p, wavelength)
        elif self.polynomial_removal is 'divide':
            p = np.polyfit(wavelength, diff, self.polynomial_order)
            diff_no_poly = diff / np.polyval(p, wavelength)
        else:
            diff_no_poly = diff

        return diff_no_poly
