import numpy as np
import scipy.constants


class RRSCrossSectionsBase(object):
    """
    Base class for RRS ring effect cross section computation. The base class should not be used directly.

    Objects derived from this base class can be used to calculate the cross section of a particular energy level
    transition within the target molecule. It can also be used to calculate cross section as a function of wavelength.

    References
    ----------
    D. Han, L. Chen, L. Su, J. Tao, S. Li, C. Yu, Y. Zhang, and B. He.  A convolution algorithm to calculate
    differential cross sections of the ring effect in the earth’s atmosphere based on rotational raman scattering.
    Science China Earth Sciences, 54(9):1407-1412, 2011.

    K. V. Chance and R. J. D. Spurr. Ring effect studies: Rayleigh scattering, including molecular parameters for
    rotational Raman scattering, and the Fraunhofer spectrum. Applied Optics, 36(21):5224-5230, 1997.
    """
    def __init__(self, excitation_wavelength_nm):
        self.excitation_wavelength = excitation_wavelength_nm
        self._b = 0.0  # rotational constant
        self._d = 0.0  # centrifugal distortion constant

        # constants
        self._h = scipy.constants.Planck
        self._c = scipy.constants.speed_of_light
        self._k = scipy.constants.Boltzmann
        self._ev = scipy.constants.electron_volt
        self._hc = 100 * self._h * self._c / self._ev  # hc in eV.cm

    @property
    def excitation_wavelength(self) -> float:
        """
        float : The wavelength of the incident photon in nm.
        """
        return self._e_wl_nm

    @excitation_wavelength.setter
    def excitation_wavelength(self, value: float):
        self._e_wl_nm = value         # nm
        self._e_wl_cm = 1e-7 * value  # cm
        self._e_wn = 1e7 / value      # cm^-1

    def _calc_partition_function(self, nmax):
        self._z = 0.0
        for n in range(nmax + 1):
            self._z += self._boltzmann_factor(n, n, 250.0)

    def calc_cross_section(self, n, n_p, j, j_p, t):
        """
        Calculate the scattered wavelength and cross section associated with a transition from state (n,j) to (n',j').

        Parameters
        ----------
        n : int
            Initial rotational angular momentum quantum number.
        n_p: int
            Final rotational angular momentum quantum number.
        j : int
            Initial total angular momentum quantum number.
        j_p : int
            Final total angular momentum quantum number.
        t : float
            Temperature in degrees Kelvin.

        Returns
        -------
        float, float
            scattered wavelength, cross section magnitude
        """
        # equation 1

        # get the polarizability anisotropy
        gamma = self._polarizability_anisotropy()

        # get the fractional population of the initial state
        f_n = self._boltzmann_factor(n, j, t) / self._z

        # get the Placzek-Teller coefficient
        b = self._placzek_teller(j, j_p)

        # get the emitted wavelength
        e = self._rotational_energy(j) - self._rotational_energy(j_p)
        wl = self._hc / e
        wl_p = self._e_wl_cm * wl / (self._e_wl_cm + wl)

        # calculate the cross section magnitude
        q = (256 * np.pi**5) / (27 * wl_p**4) * gamma**2 * f_n * b

        return 1e7 * wl_p, q

    def _polarizability_anisotropy(self):
        # equation 2
        # needs to be overridden
        return 0

    def _boltzmann_factor(self, n, j, t):
        # equation 3 * Z
        g_n = self._nuclear_spin_statistical_weight_factor(n)
        e_tot = self._rotational_energy(j)
        kt = self._k * t / self._ev  # J to eV
        return g_n * (2 * j + 1) * np.exp(-e_tot / kt)

    def _rotational_energy(self, j):
        # equation 4
        e_tot = self._hc * (self._b * (j * (j + 1)) + self._d * (j * (j + 1))**2)
        return e_tot

    def _placzek_teller(self, j, j_p):
        # equation 5
        # needs to be overridden
        return 0

    def _nuclear_spin_statistical_weight_factor(self, n):
        # needs to be overridden
        return 0.0

    def calc_cross_sections(self, nlevels, temperature):
        """
        Calculates cross section as a function of wavelength.

        Lines are not broadened at all - they are represented as perfect discrete delta functions. The returned
        wavelengths have a resolution of 0.001nm, so all line locations are rounded to the nearest 0.001nm.

        Parameters
        ----------
        nlevels : int
            The number of energy levels to consider - equal to the number of stokes lines and the number of anti-stokes
            lines calcualted.
        temperature : float
            Temperature in degrees Kelvin.

        Returns
        -------
        numpy.array, numpy.array
            Wavelengths, cross sections - cross section as a function of wavelength.
        """
        lines_wl = []
        lines_cs = []
        for i in range(nlevels):
            # stokes lines
            wl, cs = self.calc_cross_section(i, i + 2, i, i + 2, temperature)
            lines_wl.append(wl)
            lines_cs.append(cs)
            # anti-stokes lines
            wl, cs = self.calc_cross_section(i + 2, i, i + 2, i, temperature)
            lines_wl.append(wl)
            lines_cs.append(cs)

        min_wl = self._e_wl_nm - round(self._e_wl_nm - min(lines_wl), 3) - 0.01
        max_wl = self._e_wl_nm + round(max(lines_wl) - self._e_wl_nm, 3) + 0.01
        final_wl = np.arange(min_wl, max_wl + 0.001, 0.001)
        final_cs = np.zeros_like(final_wl)

        for wl, cs in zip(lines_wl, lines_cs):
            round_wl = np.round(wl, 3)
            i = int(np.round((round_wl - min_wl) * (len(final_wl) - 1) / (max_wl - min_wl)))
            final_cs[i] = cs

        return final_wl, final_cs


class RRSCrossSectionsN2(RRSCrossSectionsBase):
    """
    Class to calculate Rotationl Raman Scattering (RRS) cross sections for
    N2 molecules.

    The selection rules for N2 RRS are

     - Delta J = +/- 2
     - N = J

    where J is the total angular momentum quantum number and N is the
    rotational angular momentum quantum number.

    References
    ----------
    D. Han, L. Chen, L. Su, J. Tao, S. Li, C. Yu, Y. Zhang, and B. He.  A convolution algorithm to calculate
    differential cross sections of the ring effect in the earth’s atmosphere based on rotational raman scattering.
    Science China Earth Sciences, 54(9):1407-1412, 2011.

    K. V. Chance and R. J. D. Spurr. Ring effect studies: Rayleigh scattering, including molecular parameters for
    rotational Raman scattering, and the Fraunhofer spectrum. Applied Optics, 36(21):5224-5230, 1997.
    """
    def __init__(self, excitation_wavelength):
        super().__init__(excitation_wavelength)
        self._b = 1.989574  # rotational constant (cm^-1)
        self._d = 5.76e-6  # centrifugal distortion constant (cm^-1)
        self._calc_partition_function(40)

    def _polarizability_anisotropy(self):
        # equation 2
        return -6.01466e-25 + 2.38557e-14 / (1.86099e10 - self._e_wn**2)

    def _placzek_teller(self, j, j_p):
        # equation 5
        if j_p == j + 2:
            return 3 * (j + 1) * (j + 2) / (2 * (2 * j + 1) * (2 * j + 3))
        elif j_p == j - 2:
            return 3 * j * (j - 1) / (2 * (2 * j - 1) * (2 * j + 1))
        elif j_p == j:
            return j * (j + 1) / ((2 * j - 1) * (2 * j + 3))
        else:
            raise ValueError('invalid total angular momentum number '
                             'transition for N2: %i to %i' % (j, j_p))

    def _nuclear_spin_statistical_weight_factor(self, n):
        if n % 2 == 0:
            return 6
        else:
            return 3


class RRSCrossSectionsO2(RRSCrossSectionsBase):
    """
    Class to calculate Rotationl Raman Scattering (RRS) cross sections for
    O2 molecules.

    The selection rules for O2 RRS are:

     - Delta J = +/- 2
     - N = J

    where J is the total angular momentum quantum number and N is the
    rotational angular momentum quantum number.

    Oxygen has other relevent transitions related to a triplet state,
    but according to Dong et. al. these can be ignored for the purposes of
    DOAS retrievals, so O2 molecules are treated the same as N2 molecules.

    References
    ----------
    D. Han, L. Chen, L. Su, J. Tao, S. Li, C. Yu, Y. Zhang, and B. He.  A convolution algorithm to calculate
    differential cross sections of the ring effect in the earth’s atmosphere based on rotational raman scattering.
    Science China Earth Sciences, 54(9):1407-1412, 2011.

    K. V. Chance and R. J. D. Spurr. Ring effect studies: Rayleigh scattering, including molecular parameters for
    rotational Raman scattering, and the Fraunhofer spectrum. Applied Optics, 36(21):5224-5230, 1997.
    """
    def __init__(self, excitation_wavelength):
        super().__init__(excitation_wavelength)
        self._b = 1.4377  # rotational constant (cm^-1)
        self._d = 5e-6  # centrifugal distortion constant (cm^-1)
        self._calc_partition_function(40)

    def _polarizability_anisotropy(self):
        # equation 2
        return 7.149e-26 + 4.59364e-15 / (4.82716e9 - self._e_wn**2)

    def _placzek_teller(self, j, j_p):
        # equation 5
        if j_p == j + 2:
            return 3 * (j + 1) * (j + 2) / (2 * (2 * j + 1) * (2 * j + 3))
        elif j_p == j - 2:
            return 3 * j * (j - 1) / (2 * (2 * j - 1) * (2 * j + 1))
        elif j_p == j:
            return j * (j + 1) / ((2 * j - 1) * (2 * j + 3))
        else:
            raise ValueError('invalid total angular momentum number '
                             'transition for N2: %i to %i' % (j, j_p))

    def _nuclear_spin_statistical_weight_factor(self, n):
        if n % 2 == 0:
            return 0
        else:
            return 1
