import numpy as np
from typing import Union
from skdoas.scd.scd import SlantColumnFitterNonLinear
from scipy.interpolate import interp1d


class SlantColumnFitterOMNO2(SlantColumnFitterNonLinear):
    """
    This class carries out the spectral fitting for OMI as described in Section 2.1 of Bucsela et. al., with no
    de-striping. Absorption cross sections for NO2, O3, and H2O are fit to the measured spectra, as well as a
    Ring spectrum.

    References
    ----------
    E. J. Bucsela, N. A. Krotkov, E. A. Celarier, L. N. Lamsal, W. H. Swartz, P. K. Bhartia, K. F. Boersma, J. P.
    Veefkind, J. F. Gleason, and K. E. Pickering. A new stratospheric and tropospheric NO2 retrieval algorithm for
    nadir-viewing sateelite instruments: applications to OMI. Atmospheric Measurement Techniques, 6:2607-2626, 2013.
    """
    def __init__(self, cross_section_no2: interp1d, cross_section_o3: interp1d, cross_section_h2o: interp1d,
                 ring_spectrum: interp1d):
        super().__init__(5, [cross_section_no2, cross_section_o3, cross_section_h2o], ring_spectrum, (405., 465.))

    def fit(self, wl: np.array, rad: np.array, err: Union[np.array, None]=None, **kwargs):
        results = super().fit(wl, rad, err, **kwargs)
        results['no2_scd'] = results['scds'][0]
        results['o3_scd'] = results['scds'][1]
        results['h2o_scd'] = results['scds'][2]
        del results['scds']
        results['no2_scd_one_sigma'] = results['scds_one_sigma'][0]
        results['o3_scd_one_sigma'] = results['scds_one_sigma'][1]
        results['h2o_scd_one_sigma'] = results['scds_one_sigma'][2]
        del results['scds_one_sigma']
        return results
