import sasktran as sk
import numpy as np
from scipy.interpolate import interp1d
from scipy.optimize import curve_fit
from typing import List, Tuple, Union


class SlantColumnFitterNonLinear(object):
    """
    General class for fitting cross sections to measured spectra to derive slant column densities.

    The following parameters can be specified for the fit:

     - the order of the polynomial representing the slow-changing spectral components
     - at least one absorbing species cross sections
     - a Ring spectrum to account for inelastic scattering
     - the wavelength window to perform the fit in

    This classes uses the Levenberg-Marquardt algorithm as implemented in scipy.optimize.curve_fit.
    """
    def __init__(self, poly_degree: int, cross_sections: List[interp1d], ring_spectrum: Union[interp1d, None],
                 window: Tuple[float, float]):
        self.polynomial_degree = poly_degree
        self.cross_sections = cross_sections
        self.ring_spectrum = ring_spectrum
        self.window = window

    @property
    def polynomial_degree(self) -> int:
        """
        int : Order of the polynomial to fit to the spectrum.
        """
        return self._poly_degree

    @polynomial_degree.setter
    def polynomial_degree(self, value: int):
        self._poly_degree = int(value)

    @property
    def cross_sections(self) -> List[interp1d]:
        """
        List[interp1d] : List of interpolation objects representing cross sections for all absorbing species that will
        be included in the fit. The interpolation must be valid for all wavelengths where the spectrum is defined.
        """
        return self._cs_interp

    @cross_sections.setter
    def cross_sections(self, value: List[interp1d]):
        for cs in value:
            assert(isinstance(cs, interp1d))
        self._cs_interp = value

    @property
    def ring_spectrum(self) -> Union[interp1d, None]:
        """
        interp1d : Interpolation object representing the differential Ring spectrum to be used in the fit. If this is
        set to None then the Ring effect is not accounted for.
        """
        return self._ring_interp

    @ring_spectrum.setter
    def ring_spectrum(self, value: Union[interp1d, None]):
        if value is not None:
            assert(isinstance(value, interp1d))
            self._ring_interp = value
        else:
            self._ring_interp = None

    @property
    def window(self) -> Tuple[float, float]:
        """
        Tuple[float, float] : Wavelength window in nm where the fit should be performed. Radiances at wavelengths
        outside this window will be ignored when passed to the fit function.
        """
        return self._window

    @window.setter
    def window(self, value: Tuple[float, float]):
        self._window = (float(value[0]), float(value[1]))

    def function(self, wl: np.array, *fit_params):
        """
        The non-linear DOAS equation.

        Parameters
        ----------
        wl : numpy.array
            Wavelengths in nm to evaluate the equation at.
        *fit_params : List[float]
            The parameters that are being optimized. Should consists of the following values (in this order):

             - all polynomial coefficients in order of descending power
             - all slant column densities (in the same order as the provided cross sections)
             - the Ring amplitude (only if a Ring spectrum was provided)

        Returns
        -------
        numpy.array
            The non-linear DOAS equation evaluated at wl and fit_params. The array has shape (len(wl),).
        """
        if self._ring is None:
            assert(len(fit_params) == self.l + self.m)
            poly = np.atleast_1d(fit_params[:self.l])
            scds = np.atleast_1d(fit_params[self.l:])

            p = np.polyval(poly, wl)
            tx = np.exp(-self._cs @ scds)
            return p * tx
        else:
            assert(len(fit_params) == self.l + self.m + 1)
            poly = np.atleast_1d(fit_params[:self.l])
            scds = np.atleast_1d(fit_params[self.l:self.l + self.m])
            ring = fit_params[self.l + self.m]

            p = np.polyval(poly, wl)
            tx = np.exp(-self._cs @ scds)
            r = 1 + ring * self._ring
            return p * tx * r

    def jacobian(self, wl, *fit_params):
        """
        The Jacobian for the non-linear DOAS equation.

        Parameters
        ----------
        wl : numpy.array
            Wavelengths in nm to evaluate the equation at.
        *fit_params : List[float]
            The parameters that are being optimized. Should consists of the following values (in this order):

             - all polynomial coefficients in order of descending power
             - all slant column densities (in the same order as the provided cross sections)
             - the Ring amplitude (only if a Ring spectrum was provided)

        Returns
        -------
        numpy.array
            The jacobian of the non-linear DOAS equation evaluated at wl and fit_params. The array has shape
            (len(wl), len(fit_params)).
        """
        if self._ring is None:
            assert(len(fit_params) == self.l + self.m)
            poly = np.atleast_1d(fit_params[:self.l])
            scds = np.atleast_1d(fit_params[self.l:])

            p = np.polyval(poly, wl)
            tx = np.exp(-self._cs @ scds)
            f = p * tx

            jac = np.zeros((self.n, self.l + self.m))
            jac[:, :self.l] = wl.reshape(self.n, 1)**np.arange(self.l)[::-1].reshape(1, self.l) * tx.reshape(self.n, 1)
            jac[:, self.l:] = -f.reshape(self.n, 1) * self._cs

            return jac
        else:
            assert(len(fit_params) == self.l + self.m + 1)
            poly = np.atleast_1d(fit_params[:self.l])
            scds = np.atleast_1d(fit_params[self.l:self.l + self.m])
            ring = fit_params[self.l + self.m]

            p = np.polyval(poly, wl)
            tx = np.exp(-self._cs @ scds)
            r = 1 + ring * self._ring
            f = p * tx * r

            jac = np.zeros((self.n, self.l + self.m + 1))
            jac[:, :self.l] = wl.reshape(self.n, 1)**np.arange(self.l)[::-1].reshape(1, self.l) \
                * tx.reshape(self.n, 1) * r.reshape(self.n, 1)
            jac[:, self.l:self.l + self.m] = -f.reshape(self.n, 1) * self._cs
            jac[:, self.l + self.m] = self._ring * p * tx

            return jac

    def fit(self, wl: np.array, rad: np.array, err: Union[np.array, None]=None, **kwargs):
        """
        Performs a spectral fit on the supplied radiance data in order to retrieve the slant column densities.

        Parameters
        ----------
        wl : numpy.array
            The wavelengths in nm where the radiance is specified.
        rad : numpy.array
            The radiance spectrum to fit to.
        err : Union[numpy.array, None]
            The error associated with rad. If err is set to None, equal error is assumed across all wavelengths and
            the retrieved one sigma values will not be physically meaningful.
        **kwargs
            Keyword arguments to pass to scipy.optimize.curve_fit.

        Returns
        -------
        dict
            Dictionary containing the following entries:

             - 'popt': The optimal parameters in the order required for function and jacobian.
             - 'pcov': The covariance matrix associated with the optimal parameters.
             - 'scds': The slant column densities extracted from popt, in the order of the provided cross sections.
             - 'scds_one_sigma': The one sigma uncertainty associated with each slant column density.
             - 'ring': The retrieved Ring amplitude. Only included if a Ring spectrum was provided.
             - 'ring_one_sigma': The one sigma uncertaintly associated with the Ring amplitude. Only included if
               a Ring spectrum was provided.
             - 'chi_sq': The chi squared value for the fit. Only included if err is supplied. Values close to 1
               indicate that the fitting residual matches the error well.

        """
        window = np.logical_and(wl >= self._window[0], wl <= self._window[1])
        wl = wl[window]

        self._cs = np.vstack([c(wl) for c in self._cs_interp]).T  # n x l
        self._ring = self._ring_interp(wl) if self._ring_interp is not None else None

        self.l = self._poly_degree + 1
        self.m = self._cs.shape[1]
        self.n = len(wl)

        if 'p0' in kwargs:
            p0 = kwargs.pop('p0')
        else:
            p0 = list(np.polyfit(wl, rad, self._poly_degree))  # use polyfit as initial guess for polynomial
            p0 += [.01 / np.average(self._cs[:, i]) for i in range(self.m)]  # use od=.01 as initial guess for scds
            if self._ring is not None:
                p0 += [.01 / np.average(np.abs(self._ring))]

        popt, pcov = curve_fit(self.function, wl, rad, p0=p0, sigma=err, absolute_sigma=err is not None,
                               jac=self.jacobian, **kwargs)

        output_dict = dict()
        output_dict['popt'] = popt
        output_dict['pcov'] = pcov
        output_dict['scds'] = popt[self.l:self.l + self.m]
        output_dict['scds_one_sigma'] = np.sqrt(np.diag(pcov)[self.l:self.l + self.m])
        if self.ring_spectrum is not None:
            output_dict['ring'] = popt[-1]
            output_dict['ring_one_sigma'] = np.sqrt(pcov[-1, -1])
        if err is not None:
            output_dict['chi_sq'] = np.sum(((rad - self.function(wl, *popt)) / err) ** 2) / (self.n - self.l - self.m)
        return output_dict
