import numpy as np
import sasktran as sk
from tqdm import tqdm
from typing import List, Union

import skdoas.wf.finite_difference_wf
import skdoas.wf.wf


class WeightingFunctionBaseDO(skdoas.wf.wf.WeightingFunctionBase):
    """
    Base class for objects that calculate weighting functions using SASKTRAN's discrete ordinates (DO) engine. This
    class overrides engine-specific virtual methods and properties from WeightingFunctionBase.

    See the base class for more documentation.
    """
    def __init__(self, geometry: sk.Geometry, atmosphere: sk.Atmosphere, wf_species: Union[str, List[str]], **options):
        super().__init__(geometry, atmosphere, wf_species, **options)
        self._ref = None

    @property
    def wf_altitudes(self):
        for key in self.options:
            if key.lower() == 'wfaltitudes':
                return self.options[key]
        return np.linspace(500.0, 99500.0, 100)

    @property
    def wf_widths(self):
        for key in self.options:
            if key.lower() == 'wfwidths':
                return self.options[key]
        return np.ones(100) * 1000.0

    @property
    def reference_point(self):
        """ Reference point(s) used for the weighting function calculation. If
        self.options['averagereferencepoint'] exists and is set to True, this will return a single reference point,
        the average of the ground intersections of all lines of sight in self.geometry. Otherwise, this will return a
        list of reference points corresponding to the list of lines of sight in self.geometry. """
        refs_geodetic = [los.ground_intersection() for los in self.geometry.lines_of_sight]
        refs_latlonalt = [np.array([r.latitude, r.longitude, 0.0, los.mjd]) for r, los in zip(refs_geodetic, self.geometry.lines_of_sight)]
        for key in self.options:
            if key.lower() == 'averagereferencepoint':
                if self.options[key]:
                    refs_xyz = np.reshape([ref.location for ref in refs_geodetic], (-1, 3))
                    ref_xyz = np.average(refs_xyz, axis=0)
                    ref = sk.Geodetic()
                    ref.from_xyz(ref_xyz)
                    mjd = np.average([los.mjd for los in self.geometry.lines_of_sight])
                    return np.array([ref.latitude, ref.longitude, 0.0, mjd])
        return refs_latlonalt

    @property
    def atmosphere_grid(self) -> np.array:
        for key in self.options:
            if key.lower() == 'altitudegrid':
                return self.options[key]
        return np.linspace(0, 1e5, 201)

    @property
    def surface_height(self) -> float:
        return self.atmosphere_grid[0]

    @property
    def top_of_atmosphere(self) -> float:
        return self.atmosphere_grid[-1]

    def append_atmosphere_grid(self, values: np.array):
        # save the appropriate options
        grid_names = ['altitudegrid']
        pop = [name for name in self.options.keys() if name.lower() in grid_names]
        self._temp_options = {key: self.options.pop(key) for key in pop}

        # create the new manual grids
        grid = np.linspace(0, 1e5, 201)  # default
        for key, value in self._temp_options.items():
            if key.lower() == grid_names[0]:
                grid = value

        # add, sort and filter the new altitudes
        grid = np.append(grid, values)
        grid[grid < self.surface_height] = self.surface_height
        grid = np.unique(grid)

        # set the manual grids
        self.options[grid_names[0]] = grid

    def restore_atmosphere_grid(self):
        if hasattr(self, '_temp_options'):
            self.options.pop('altitudegrid', None)
            for key in self._temp_options:
                self.options[key] = self._temp_options[key]
            del self._temp_options

    def radiance(self, wl: List[float], last_wf=False):
        if last_wf is True:
            return self._rad
        else:
            temp_spec = self.atmosphere.wf_species
            self.atmosphere.wf_species = None
            self._engine = sk.EngineDO(self.geometry, self.atmosphere, wl, self.options)
            rad = self._engine.calculate_radiance(self.output_format)
            self.atmosphere.wf_species = temp_spec
            return rad


class WeightingFunctionDO(WeightingFunctionBaseDO):
    """
    Class to calculate weighting functions using the built-in method from SASKTRAN's discrete ordinates (DO) engine.

    See base classes for more documentation.
    """
    def __init__(self, geometry: sk.Geometry, atmosphere: sk.Atmosphere, wf_species: Union[str, List[str]],
                 **options):
        super().__init__(geometry, atmosphere, wf_species, **options)

    def __call__(self, wl: List[float]):
        # hold a temporary copy of the atmosphere's wf_species
        temp_spec = self.atmosphere.wf_species

        # Setup engine and do calculation
        self.atmosphere.wf_species = self.wf_species
        self._engine = sk.EngineDO(self.geometry, self.atmosphere, wl, self.options)
        if self.output_format == 'numpy':
            # calculate radiance and weighting functions
            self._rad, wf = self._engine.calculate_radiance('numpy')

            # Get box altitudes (wf altitudes)
            box_alts = self._engine.options.get('wfaltitudes', np.linspace(500, 99500, 100))

            # Get dimensions
            nlos = len(self.geometry.lines_of_sight)
            nwfs = len(self.wf_species)
            nwl = len(wl)
            nbox = len(box_alts)

            # Reshape wf to account for multiple species
            wf = wf.reshape((nwl, nlos, nwfs, nbox))  # shape: (w, l, s, b)
        elif self.output_format == 'xarray':
            wf = self._engine.calculate_radiance('xarray')
            specs = [f'wf_{spec}' for spec in np.atleast_1d(self.wf_species)]
            self._rad = wf.drop(specs)
            # self._rad = wf.radiance.values
        else:
            raise ValueError('output_format must be \'numpy\' or \'xarray\'')

        # restore wf_species
        self.atmosphere.wf_species = temp_spec

        return wf


class WeightingFunctionFiniteDifferenceDO(WeightingFunctionBaseDO,
                                          skdoas.wf.finite_difference_wf.WeightingFunctionFiniteDifferenceBase):
    """
    Class to calculate weighting functions with SASKTRAN's discrete ordinates (DO) engine using finite difference
    approximations. Use of this class is not recommended, as WeightingFunctionDO produces equivalent results in a
    fraction of the time.

    See base classes for more documentation.
    """
    def __init__(self, geometry: sk.Geometry, atmosphere: sk.Atmosphere, wf_species: Union[str, List[str]],
                 method: str='forward', step: Union[float, List[float]]=0.01, relative_step=True, minimum_step=1e8,
                 progress_bar: Union[bool, tqdm]=False, **options):
        options = dict(averagereferencepoint=True, **options)
        WeightingFunctionBaseDO.__init__(self, geometry, atmosphere, wf_species, **options)
        skdoas.wf.finite_difference_wf.WeightingFunctionFiniteDifferenceBase.__init__(
            self, geometry, atmosphere, wf_species, method, step, relative_step, minimum_step, progress_bar, **options)
        self._ref = None
