import numpy as np
import sasktran as sk

from typing import List, Union

import skdoas.wf.finite_difference_wf
import skdoas.wf.wf


class WeightingFunctionBaseMC(skdoas.wf.wf.WeightingFunctionBase):
    """
    Base class for objects that calculate weighting functions using SASKTRAN's Monte Carlo (MC) engine. This
    class overrides engine-specific virtual methods and properties from WeightingFunctionBase.

    See the base class for more documentation.
    """
    def __init__(self, geometry: sk.Geometry, atmosphere: sk.Atmosphere, wf_species: Union[str, List[str]], **options):
        super().__init__(geometry, atmosphere, wf_species, **options)

    @property
    def reference_point(self):
        """
        numpy.array : Reference point used for the weighting function calculation. This will return
        self.options['referencepoint'] if it exists, else it will return the default selected by the MC engine.
        """
        for key in self.options.keys():
            if key.lower() is 'referencepoint':
                return self.options[key]

        engine = sk.Engine('mc', self.geometry, self.atmosphere, [500.0], self.options)
        engine._add_lines_of_sight()
        engine._initialize_model()
        return engine.model_parameters['referencepoint']

    @property
    def atmosphere_grid(self) -> np.array:
        for key in self.options:
            if key.lower() == 'manualopticalheights':
                return self.options[key]
        return np.linspace(0, 1e5, 201)

    @property
    def atmosphere_grid_handle(self) -> str:
        return 'manualopticalheights'

    @property
    def surface_height(self) -> float:
        for key in self.options:
            if key.lower() == 'surfaceheight':
                return self.options[key]
        return 0.0

    @property
    def top_of_atmosphere(self) -> float:
        for key in self.options:
            if key.lower() == 'toaheight':
                return self.options[key]
        return 1e5

    def append_atmosphere_grid(self, values: np.array):
        # save the appropriate options
        grid_names = ['setnumraytracingshells', 'setsolartablealtitudedelta', 'manualraytracingshells',
                      'manualsolarraytracingshells', 'manualopticalheights']
        pop = [name for name in self.options.keys() if name.lower() in grid_names]
        self._temp_options = {key: self.options.pop(key) for key in pop}

        # get ground and top of atmosphere altitudes
        surface, toa = 0.0, 1e5
        for key, value in self.options.items():
            if key.lower() == 'surfaceheight':
                surface = value
            elif key.lower() == 'toaheight':
                toa = np.ceil(value)

        # create the new manual grids
        shells = np.linspace(0.0, 1e5, 101)  # default
        solar = np.linspace(0.0, 1e5, 101)  # default
        optical = np.linspace(0.0, 1e5, 201)  # default
        for key, value in self._temp_options.items():
            if key.lower() == grid_names[0]:  #
                shells = np.linspace(surface, toa, value + 1)
            elif key.lower() == grid_names[1]:
                solar = np.arange(surface, toa, value)
        for key, value in self._temp_options.items():  # manual grids override spacing/numshells in MC
            if key.lower() == grid_names[2]:
                shells = value
            elif key.lower() == grid_names[3]:
                solar = value
            elif key.lower() == grid_names[4]:
                optical = value

        # add, sort and filter the new altitudes
        shells = np.append(shells, values)
        shells[shells < surface] = surface
        shells[shells > toa] = toa
        shells = np.unique(shells)

        solar = np.append(solar, values)
        solar[solar < surface] = surface
        solar[solar > toa] = toa
        solar = np.unique(solar)

        optical = np.append(optical, values)
        optical[optical < surface] = surface
        optical[optical > toa] = toa
        optical = np.unique(optical)

        # set the manual grids
        self.options[grid_names[2]] = shells
        self.options[grid_names[3]] = solar
        self.options[grid_names[4]] = optical

    def restore_atmosphere_grid(self):
        if hasattr(self, '_temp_options'):
            self.options.pop('manualraytracingshells', None)
            self.options.pop('manualsolarraytracingshells', None)
            self.options.pop('manualopticalheights', None)
            for key in self._temp_options:
                self.options[key] = self._temp_options[key]
            del self._temp_options


class WeightingFunctionNullMC(WeightingFunctionBaseMC):
    """
    This class does not actually calculate any weighting functions; using the __call__ method will raise an exception.
    Its only use is to keep air mass factor class structure consistent, as HR and DO use weighting functions to
    calculate air mass factors but MC does not. This class can calculate radiances and hold various settings,
    functions which the other weighting function classes provide in addition to weighting function calculation.

    See base classes for more documentation.
    """
    def __init__(self, geometry: sk.Geometry, atmosphere: sk.Atmosphere, **options):
        super().__init__(geometry, atmosphere, [], **options)

    def __call__(self, wl: List[float]):
        """
        Calling this object will raise a NotImplementedError exception.
        """
        raise NotImplementedError()

    def radiance(self, wl: List[float], last_wf=False):
        if last_wf is True:
            return self._rad
        else:
            self._engine = sk.Engine('MC', self.geometry, self.atmosphere, wl, self.options)
            self._rad = self._engine.calculate_radiance(self.output_format)
            return self._rad
