import sasktran as sk
import numpy as np

from typing import List, Union


class WeightingFunctionBase(object):
    """
    Base class for objects which calculate weighting functions using SASKTRAN.
    """
    def __init__(self, geometry: sk.Geometry, atmosphere: sk.Atmosphere, wf_species: Union[str, List[str]], **options):
        """
        Parameters
        ----------
        geometry: sk.Geometry
            Geometry object from sasktran containing information about the lines of sight and the sun.
        atmosphere: sk.Atmosphere
            Atmosphere object from sasktran containing information about the atmosphere. If atmosphere.wf_species
            is set, it will be overruled by wf_species.
        wf_species: str, List[str]
            List of species corresponding to the names of species in atmosphere for which weighting functions will be
            calculated.
        **options
            Options that are passed to the engine.
        """
        self.atmosphere = atmosphere
        self.geometry = geometry
        self.wf_species = wf_species
        self.options = options
        self.output_format = 'xarray'
        self._rad = None
        self._engine = None

    @property
    def wf_species(self) -> List[str]:
        """
        List[str] : List of atmospheric species to calculate weighting functions for.
        """
        return self._wf_species

    @wf_species.setter
    def wf_species(self, species: Union[str, List[str]]):
        self._wf_species = list(np.atleast_1d(species))

    @property
    def atmosphere(self) -> sk.Atmosphere:
        """
        sasktran.Atmosphere : Atmosphere definition for the calculation.
        """
        return self._atmosphere

    @atmosphere.setter
    def atmosphere(self, atmosphere: sk.Atmosphere()):
        self._atmosphere = atmosphere

    @property
    def geometry(self) -> sk.Geometry:
        """
        sasktran.Geometry : Definitions of line(s) of sight and sun position.
        """
        return self._geometry

    @geometry.setter
    def geometry(self, geometry: sk.Geometry()):
        self._geometry = geometry

    @property
    def options(self) -> dict:
        """
        dict : Options to pass to the sasktran engine.
        """
        return self._options

    @options.setter
    def options(self, options: dict):
        self._options = options

    @property
    def engine(self) -> sk.Engine:
        """
        sasktran.Engine : Engine used for the calculation.
        """
        return self._engine

    @property
    def output_format(self) -> str:
        """
        str : Output format, must be 'numpy' or 'xarray'.
        """
        return self._out

    @output_format.setter
    def output_format(self, value):
        if value.lower() == 'numpy':
            self._out = 'numpy'
        elif value.lower() == 'xarray':
            self._out = 'xarray'
        else:
            raise ValueError('output_format must be \'numpy\' or \'xarray\'')

    @property
    def wf_altitudes(self) -> np.array:
        """
        numpy.array : Weighting function altitudes taken from self.options if present. Must be overridden in an
        engine-specific subclass.
        """
        raise NotImplementedError()

    @property
    def wf_widths(self) -> np.array:
        """
        numpy.array : Weighting function widths taken from self.options if present. Must be overridden in an
        engine-specific subclass.
        """
        raise NotImplementedError()

    @property
    def reference_point(self) -> np.array:
        """
        numpy.array : Reference point(s) in form np.array([latitude, longitude, altitude, mjd]). Must be overridden in
        an engine-specific subclass.
        """
        raise NotImplementedError()

    @property
    def atmosphere_grid(self) -> np.array:
        """
        numpy.array : Altitudes where atmospheric profiles are defined in metres. Must be overridden in an
        engine-specific subclass.
        """
        raise NotImplementedError()

    @property
    def top_of_atmosphere(self) -> float:
        """
        float : Top of atmosphere in metres. Must be overridden in an engine-specific subclass.
        """
        raise NotImplementedError()

    @property
    def surface_height(self) -> float:
        """
        float : Surface height in metres. Must be overridden in an engine-specific subclass.
        """
        raise NotImplementedError()

    def append_atmosphere_grid(self, values: np.array):
        """
        Adds values to the atmospheric grid set in self.options (adds to the default value if it is not set). This
        method should always be undone by a subsequent call to restore_atmosphere_grid. Must be overridden in an
        engine-specific subclass.
        """
        raise NotImplementedError()

    def restore_atmosphere_grid(self):
        """
        Restores the atmospheric grid to its original value after it has been altered by append_atmosphere_grid.
        Must be overridden in an engine-specific subclass.
        """
        raise NotImplementedError()

    def radiance(self, wl: List[float], last_wf=False):
        """
        Calculates or fetches radiance values. Built-in weighting function methods should never be used when the
        radiance is calculated here (these can be used by __call__ instead). Must be overridden in an engine-specific
        subclass.

        Parameters
        ----------
        wl: List[float]
            Wavelengths in nm to calculate radiances at.
        last_wf: bool, False
            If True, the radiance from the last wf calculation is retrieved (this may not correspond to the given wl).
            If False, the radiance is calculated.

        Returns
        -------

        numpy.ndarray **or** xarray.Dataset
            If output_format='numpy':
            Radiances are returned in an array with shape (len(wavelengths), len(lines of sight)).

            If output_format='xarray':
            Radiances are returned in an xarray Dataset.
        """
        raise NotImplementedError()

    def __call__(self, wl: List[float]):
        """
        Calculates weighting functions. Must be overridden in an engine-specific subclass.

        Parameters
        ----------
        wl: List[float]
            Wavelengths in nm to calculate weighting functions at.

        Returns
        -------

        numpy.ndarray **or** xarray.Dataset
            If output_format='numpy':
            Weighting Functions are returned in an array with shape
            (len(wavelengths), len(lines of sight), len(wf species), len(wf altitudes)).

            If output_format='xarray':
            Weighting Functions are returned in an xarray Dataset.
        """
        raise NotImplementedError()
